﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;


namespace bvctr {
	public class PenisCollision : MVRScript {
		private List<GameObject> _debugObjects = new List<GameObject>();

		/*
		* Scale value to produce estimated visual colliders.
		* Without this, copying rigidbody colliders transform results in objects larger than a default person atom.
		*/
		private float _estimatedScale = 0.08f;

		// Controllers
		protected FreeControllerV3 contPenisBase;
		protected FreeControllerV3 contPenisMid;
		protected FreeControllerV3 contPenisTip;

		// Rigidbodies
		protected Rigidbody rbPenisBase;
		protected Rigidbody rbPenisMid;
		protected Rigidbody rbPenisTip;

		// UI
		protected JSONStorableBool isBaseEnabled;
		protected JSONStorableBool isDebugEnabled;
		protected JSONStorableBool isMidEnabled;
		protected JSONStorableBool isTipEnabled;
		protected JSONStorableFloat colliderScale = new JSONStorableFloat("Scale", 1f, 0.01f, 1f, true);



		private void DebugClear(bool restart = false) {
			_debugObjects.ForEach((GameObject obj) => {
				Destroy(obj, 0f);
			});

			if (restart) {
				DebugStart();
			}
		}

		/**
		* Add estimated visualization for colliders.
		*/
		private void DebugStart() {
			Dictionary<string, Rigidbody> colliderMapping = new Dictionary<string, Rigidbody>();
			Component[] comps = rbPenisBase.GetComponentsInChildren(typeof(CapsuleCollider));

			colliderMapping.Add("AutoColliderGen1Hard", rbPenisBase);
			colliderMapping.Add("AutoColliderGen2Hard", rbPenisMid);
			colliderMapping.Add("AutoColliderGen3aHard", rbPenisTip);
			colliderMapping.Add("AutoColliderGen3bHard", rbPenisTip);

			foreach (Component c in comps) {
				GameObject cap = GameObject.CreatePrimitive(PrimitiveType.Capsule);
				cap.name = $"{c.name}_debug";
				Rigidbody rb = new Rigidbody();
				Renderer renderer = cap.GetComponent<Renderer>();
				renderer.material.shader = Shader.Find("Transparent/Diffuse");
				renderer.material.color = new Color(0f, 0f, 1f, 0.2f);

				// Set color
				if (colliderMapping.ContainsKey(c.name)) {
					if (!colliderMapping[c.name].detectCollisions) {
						renderer.material.color = new Color(1f, 0f, 0f, 0.2f);
					}
				}

				// Remove physics
				rb = cap.AddComponent<Rigidbody>();
				rb.detectCollisions = false;
				rb.useGravity = false;

				_debugObjects.Add(cap);

				// Set transform
				cap.transform.SetParent(c.transform);
				cap.transform.localScale = c.transform.localScale * _estimatedScale;
				cap.transform.localPosition = Vector3.zero;
				cap.transform.rotation = c.transform.rotation;
			}
		}

		/**
		* Transform scale of rigidbody colliders.
		*/
		private void ScaleColliders(float scale) {
			Component[] comps = rbPenisBase.GetComponentsInChildren(typeof(CapsuleCollider));

			foreach (Component c in comps) {
				CapsuleCollider col = c as CapsuleCollider;
				Vector3 newScale = new Vector3(1f, scale * 0.9f, 1f) * colliderScale.val;

				//if (col.direction == 2) {
				//	newScale = new Vector3(scale, 1f, 1f) * colliderScale.val;
				//}

				c.transform.localScale = newScale;
			}

			if (isDebugEnabled.val) {
				DebugClear(true);

				//isDebugEnabled.valNoCallback = false;
			}
		}

		/**
		* Turn rigidbody collision on or off.
		*/
		private void ToggleCollision(Rigidbody rb, bool state) {
			rb.detectCollisions = state;

			if (isDebugEnabled.val) {
				DebugClear(true);
			}
		}



		public IEnumerator DeferredInit() {
			// yield return new WaitForEndOfFrame();

			int waited = 0;

			while (waited < 6) {
				yield return new WaitForSeconds(1);
				// yield return new WaitForFixedUpdate();

				waited++;
			}

			// Apply current settings
			ToggleCollision(rbPenisBase, isBaseEnabled.val);
			ToggleCollision(rbPenisMid, isMidEnabled.val);
			ToggleCollision(rbPenisTip, isTipEnabled.val);

			containingAtom.RestoreFromLast(this);
		}

		public override void Init() {
			try {
				// Check plugin is added to male Person Atom
				if (containingAtom.type != "Person") {
					SuperController.LogMessage("PenisCollision plugin needs to be added to a Person atom.");
					return;
				}

				// Init rigidbodies
				rbPenisBase = containingAtom.rigidbodies.Where((rb) => rb.name == "Gen1").First();
				rbPenisMid = containingAtom.rigidbodies.Where((rb) => rb.name == "Gen2").First();
				rbPenisTip = containingAtom.rigidbodies.Where((rb) => rb.name == "Gen3").First();

				CreateSpacer().height = 40f;

				isBaseEnabled = new JSONStorableBool("PenisBase collision", true);
				isBaseEnabled.setCallbackFunction = (bool val) => {
					ToggleCollision(rbPenisBase, val);
				};
				RegisterBool(isBaseEnabled);
				CreateToggle(isBaseEnabled);

				isMidEnabled = new JSONStorableBool("PenisMid collision", true);
				isMidEnabled.setCallbackFunction = (bool val) => {
					ToggleCollision(rbPenisMid, val);
				};
				RegisterBool(isMidEnabled);
				CreateToggle(isMidEnabled);

				isTipEnabled = new JSONStorableBool("PenisTip collision", true);
				isTipEnabled.setCallbackFunction = (bool val) => {
					 ToggleCollision(rbPenisTip, val);
				 };
				RegisterBool(isTipEnabled);
				CreateToggle(isTipEnabled);

				// Debug toggle
				isDebugEnabled = new JSONStorableBool("Visualize", false);
				isDebugEnabled.setCallbackFunction = (bool val) => {
					if (val) {
						DebugStart();
					} else {
						DebugClear();
					}
				};
				CreateToggle(isDebugEnabled, true);

				colliderScale.setCallbackFunction = (float val) => {
					ScaleColliders(val);
				};
				RegisterFloat(colliderScale);
				CreateSlider(colliderScale, true);

				StartCoroutine(DeferredInit());
			} catch (Exception e) {
				SuperController.LogError("PenisCollision.cs Exception: " + e);
			}
		}

		public void OnDestroy() {
			DebugClear();

			// Reset collider scales
			ScaleColliders(1f);

			// Enable all collision
			ToggleCollision(rbPenisBase, true);
			ToggleCollision(rbPenisMid, true);
			ToggleCollision(rbPenisTip, true);
		}
	}
}
